from charmhelpers.core import templating
from charmhelpers.core import host


SERVICES = {}


def register(services, templates_dir=None):
    """
    Register a list of services, given their definitions.

    Service definitions are dicts in the following formats:

        {
            "service": <service name>,
            "templates": [ {
                'source': <source template to render>,
                'target': <render target of template>,
                'file_properties': <dict of file properties, see host.write_file>,
                'contexts': <list of context generators>,
            } ],
            'complete': <list of actions to call when all templates are complete>,
            'stop': <list of actions to call when stopping the service>,
        }

    If 'source' is omitted, it will default to ``basename(target).j2``.

    If 'target' is omitted, it will default to ``/etc/init/<service name>.conf``
    (i.e., an Upstart service of the given name).

    If 'complete' is omitted, it defaults to `[host.service_restart]` (i.e.,
    restarting the service via Upstart).

    If 'stop' is ommitted, it defaults to `[host.service_stop]` (i.e., stopping
    the service via Upstart).

    The `templates_dir` param indicates where to find template source files, and
    defaults to `$CHARM_DIR/templates`.


    Examples:

    The following registers an Upstart service called bingod that depends on
    a mongodb relation and which runs a custom `db_migrate` function prior to
    restarting the service, and a Monit watcher to monitor the bingod service.

        >>> services.register([
        ...     {
        ...         'service': 'bingod',
        ...         'templates': [
        ...             {'source': 'bingod.conf.j2'},
        ...             {'target': '/etc/bingod.ini',
        ...              'file_properties': {'owner': 'bingo', 'perms': 0400},
        ...              'contexts': [MongoRelationContext()]},
        ...         ],
        ...     },
        ...     {
        ...         'service': 'referee',
        ...         'templates': [{'target': '/etc/monit/conf.d/referee.cfg'}],
        ...         'complete': [monit_reload],
        ...         'stop': [],
        ...     },
        ... ])
    """
    for service in services:
        service_name = service['service']
        service.setdefault('templates_dir', templates_dir)
        for template in service['templates']:
            template.setdefault('target', '/etc/init/{}.conf'.format(service_name))
        SERVICES[service_name] = service


def reconfigure_services(*service_names):
    """
    Update all files for one or more registered services, and,
    if ready, optionally restart them.

    If no service names are given, reconfigures all registered services.
    """
    for service_name in service_names or SERVICES.keys():
        service = _get_service(service_name)
        complete = templating.render(service['templates'], service['templates_dir'])
        if complete:
            for action in service.get('complete', [host.service_restart]):
                action(service_name)


def stop_services(*service_names):
    """
    Stop one or more registered services, by name.

    If no service names are given, stops all registered services.
    """
    for service_name in service_names or SERVICES.keys():
        service = _get_service(service_name)
        complete = templating.render(service['templates'], service['templates_dir'])
        if complete:
            for action in service.get('stop', [host.service_stop]):
                action(service_name)


def _get_service(service_name):
    """
    Given the name of a registered service, return a ServiceManager
    instance for the type of that service.
    """
    service = SERVICES.get(service_name)
    if not service or service['service'] != service_name:
        raise KeyError('Service not registered: %s' % service_name)
    return service
