import os
import yaml

from charmhelpers.core import host
from charmhelpers.core import hookenv


class ContextGenerator(object):
    """
    Base interface for template context container generators.

    A template context is a dictionary that contains data needed to populate
    the template.  The generator instance should produce the context when
    called (without arguments) by collecting information from juju (config-get,
    relation-get, etc), the system, or whatever other sources are appropriate.

    A context generator should only return any values if it has enough information
    to provide all of its values.  Any context that is missing data is considered
    incomplete and will cause that template to not render until it has all of its
    necessary data.

    The template may receive several contexts, which will be merged together,
    so care should be taken in the key names.
    """
    def __call__(self):
        raise NotImplementedError


class StorableContext(object):
    """
    A mixin for persisting a context to disk.
    """
    def store_context(self, file_name, config_data):
        with open(file_name, 'w') as file_stream:
            yaml.dump(config_data, file_stream)

    def read_context(self, file_name):
        with open(file_name, 'r') as file_stream:
            data = yaml.load(file_stream)
            if not data:
                raise OSError("%s is empty" % file_name)
            return data


class ConfigContext(ContextGenerator):
    """
    A context generator that generates a context containing all of the
    juju config values.
    """
    def __call__(self):
        return hookenv.config()


class RelationContext(ContextGenerator):
    """
    Base class for a context generator that gets relation data from juju.

    Subclasses must provide `interface`, which is the interface type of interest,
    and `required_keys`, which is the set of keys required for the relation to
    be considered complete.  The first relation for the interface that is complete
    will be used to populate the data for template.

    The generated context will be namespaced under the interface type, to prevent
    potential naming conflicts.
    """
    interface = None
    required_keys = []

    def __call__(self):
        if not hookenv.relation_ids(self.interface):
            return {}

        ctx = {}
        for rid in hookenv.relation_ids(self.interface):
            for unit in hookenv.related_units(rid):
                reldata = hookenv.relation_get(rid=rid, unit=unit)
                required = set(self.required_keys)
                if set(reldata.keys()).issuperset(required):
                    ns = ctx.setdefault(self.interface, {})
                    for k, v in reldata.items():
                        ns[k] = v
                    return ctx

        return {}


class StaticContext(ContextGenerator):
    def __init__(self, data):
        self.data = data

    def __call__(self):
        return self.data


def _collect_contexts(context_providers):
    """
    Helper function to collect and merge contexts from a list of providers.

    If any of the contexts are incomplete (i.e., they return an empty dict),
    the template is considered incomplete and will not render.

    If there are no context_providers given, then it is automatically
    considered complete and an empty context (`{}`) is returned.
    """
    if not context_providers:
        return {}  # special case: no contexts at all is always complete
    ctx = {}
    for provider in context_providers:
        c = provider()
        if not c:
            return False
        ctx.update(c)
    return ctx


def render(template_definitions, templates_dir=None):
    """
    Render one or more templates, given a list of template definitions.

    The template definitions should be dicts with the keys: `source`, `target`,
    `file_properties`, and `contexts`.

    The `source` path, if not absolute, is relative to the `templates_dir`
    given when the rendered was created.  If `source` is not provided
    for a template the `template_dir` will be consulted for
    ``basename(target).j2``.

    The `target` path should be absolute.

    The `file_properties` should be a dict optionally containing
    `owner`, `group`, or `perms` options, to be passed to `write_file`.

    The `contexts` should be a list containing zero or more ContextGenerators.

    The `template_dir` defaults to `$CHARM_DIR/templates`

    Returns True if all of the templates were "complete" (i.e., the context
    generators were able to collect the information needed to render the
    template) and were rendered.

    Note: If a template has no contexts listed, it is automatically considered
    complete.
    """
    # lazy import jinja2 in case templating is needed in install hook
    from jinja2 import FileSystemLoader, Environment, exceptions
    all_complete = True
    if templates_dir is None:
        templates_dir = os.path.join(hookenv.charm_dir(), 'templates')
    loader = Environment(loader=FileSystemLoader(templates_dir))
    for tmpl in template_definitions:
        ctx = _collect_contexts(tmpl.get('contexts', []))
        if ctx is False:
            all_complete = False
            continue
        try:
            source = tmpl.get('source', os.path.basename(tmpl['target'])+'.j2')
            template = loader.get_template(source)
        except exceptions.TemplateNotFound as e:
            hookenv.log('Could not load template %s from %s.' %
                        (tmpl['source'], templates_dir),
                        level=hookenv.ERROR)
            raise e
        content = template.render(ctx)
        host.mkdir(os.path.dirname(tmpl['target']))
        host.write_file(tmpl['target'], content, **tmpl.get('file_properties', {}))
    return all_complete
